import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

import java.util.*;

/**
 * Class encapsulating state of a node in the graph we are drawing.  For
 * simplicity, we just extend ZVisualLeaf, so that we can keep individual
 * VGroupNode objects directly in the Jazz Scenegraph.
 */
public class VGroupNode {
    // camera magnification at which a dataguide node is replaced with a group node
    static public final int    MAX_SAMPLE_SIZE = 2; 
    // maximum number of data nodes displayed in a group node
    // preferably a square number

    static private final double NODE_HEIGHT = VDataguide.NODE_HEIGHT;
    static private final double NODE_WIDTH  = VDataguide.NODE_WIDTH;

    static private final double PTR_BOX_NODE_WIDTH_RATIO = 8.0;
    static private final double PTR_BOX_NODE_HEIGHT_RATIO = 6.0;

    static public final double INTER_NODE_SPACING_X = VDataguide.INTER_NODE_SPACING_X;
    static public final double INTER_NODE_SPACING_Y = VDataguide.INTER_NODE_SPACING_Y;
    static public final double ABS_PEN_WIDTH = VDataguide.ABS_PEN_WIDTH;

    private int _start_data_node_num = 0;
    private int _identifier = 0;
    private int _num_data_nodes_in_group = 0;
    private int _dg_node_id = 0;

    /**
     * constructor.
     *
     * @param dg_node_id_ index for dg node of this group
     * @param my_canvas_ canvas onto which to draw this group
     * @param node_ Jazz object
     * @param sticky_node_ deprecated
     * @param bounding_box_ bounding box within which to draw the group
     * @param node_label_string_ group's label
     * @param dataguide_ dataguide
     * @param datagraph_ data graph
     * @param offsets_vector_ offsets of data nodes in this group
     * @param identifier_ identifier for this group node (unique among all groups that belong to a dg node)
     * @param avg_char_width_ average width of a character at the current magnification level
     * @param avg_char_height_ average height of a character at the current magnification level
     * @param font_size_ size of font readable at the current magnification level
     * @param camera_mag_ current magnification level
     * @param sample_size_ size of a group
     * @param click_active_ has a node been clicked
     * @param clicked_offset_ offset of the data node that was clicked
     * @param curr_dg_node_ current dg node
     * @param colors_ Vector of colors 
     * @param summary_ should summary view be shown?
     * @param timer_thread_ timer thread
     */
   
    public VGroupNode(int dg_node_id_, ZCanvas my_canvas_, VDataguide node_, ZVisualLeaf sticky_node_, 
		      Rectangle2D bounding_box_, String node_label_string_, Dataguide dataguide_, 
		      DataGraph datagraph_, FileVector offsets_vector_, int identifier_, 
		      double avg_char_width_, double avg_char_height_, 
		      int font_size_, double camera_mag_, int sample_size_, 
		      boolean click_active_, long clicked_offset_, DataguideNode curr_dg_node_,
		      Vector colors_, boolean summary_, TimerThread timer_thread_, 
		      Hashtable dgnodeindex_histogram_, boolean partial_) {
	// compute left top corner of bounding box
	double bounding_box_ltx = bounding_box_.getX();
	double bounding_box_lty = bounding_box_.getY();
	boolean summary = summary_;
	if(summary) {
	    try {
		double inter_node_spacing_x = INTER_NODE_SPACING_X/camera_mag_; 
		double inter_node_spacing_y = INTER_NODE_SPACING_Y/camera_mag_; 
		double curr_grid_pos_x = bounding_box_.getX() + inter_node_spacing_x;
		double curr_grid_pos_y = bounding_box_.getY() + inter_node_spacing_y;
		double curr_grid_width = bounding_box_.getWidth() - 2*inter_node_spacing_x;
		double curr_grid_height = bounding_box_.getHeight() - 2*inter_node_spacing_y;
		Rectangle2D bounding_box = new Rectangle2D.Double(curr_grid_pos_x, curr_grid_pos_y,
								  curr_grid_width, curr_grid_height);
		Histogram h = null;
		if(dgnodeindex_histogram_.containsKey(new Integer(dg_node_id_))) {
		    h = (Histogram)(dgnodeindex_histogram_.get(new Integer(dg_node_id_)));
		    if(h.getNumBuckets() != sample_size_) {
			h.setNumBuckets(sample_size_);
			h.index(datagraph_, offsets_vector_, 0);
			if(!h.processingComplete()) timer_thread_.queueEvent(h);
		    }
		}
		else {
		    System.out.println("Creating histogram for node "+curr_dg_node_.getDGIndex());
		    System.out.println("dg_node_id_ = "+dg_node_id_);
		    h = new Histogram(curr_dg_node_);
		    h.setNumBuckets(sample_size_);
		    h.index(datagraph_, offsets_vector_, 0);
		    dgnodeindex_histogram_.put(new Integer(dg_node_id_), h);
		    if(!h.processingComplete()) timer_thread_.queueEvent(h);
		}
		Vector visual_components = h.display(bounding_box, camera_mag_, colors_, avg_char_height_, font_size_);
		Enumeration e = visual_components.elements();
		while(e.hasMoreElements()) {
		    node_.addVisualComponent((ZVisualComponent)(e.nextElement()));
		}
	    }
	    catch(NumberFormatException nfe_) {
		summary = false;
	    }
	}

	if(!summary) {
	    if(offsets_vector_ == null) {
		System.out.println("Terrible; check VDataguide");
		System.exit(0);
	    }
	    _dg_node_id = dg_node_id_;
	    _identifier = identifier_;
	    _start_data_node_num = _identifier*sample_size_; //VGroupNode.MAX_SAMPLE_SIZE;
	    
	    /* create a grid of rectangles corresponding to data nodes in this group */
	    
	    // compute the number of rows and cols in the grid  
	    // Explain this -- Thomas
	    int num_rem_data_nodes = offsets_vector_.size()-_start_data_node_num;
	    if(MyFrame._debug_mode > 5) {
		System.out.println("No. of remaining data nodes="+num_rem_data_nodes);
	    }
	    _num_data_nodes_in_group = sample_size_; //VGroupNode.MAX_SAMPLE_SIZE;
	    if(num_rem_data_nodes < _num_data_nodes_in_group) {
		_num_data_nodes_in_group = num_rem_data_nodes;
	    }
	    num_rem_data_nodes = num_rem_data_nodes-_num_data_nodes_in_group;
	    
	    if(MyFrame._debug_mode > 5) {
		System.out.println("No. of data nodes in group="+_num_data_nodes_in_group);
	    }
	    
	    int num_rows = (int)(Math.ceil(Math.sqrt((double)_num_data_nodes_in_group)));
	    int num_cols = (int)Math.ceil((double)_num_data_nodes_in_group/(double)num_rows);
	    if(MyFrame._debug_mode > 5) {
		System.out.println("No. of rows="+num_rows);
		System.out.println("No. of cols="+num_cols);
	    }
	    
	    double inter_node_spacing_x = INTER_NODE_SPACING_X/camera_mag_; 
	    double inter_node_spacing_y = INTER_NODE_SPACING_Y/camera_mag_; 
	    double rectangle_width = (bounding_box_.getWidth()-inter_node_spacing_x*(num_cols+1))/(num_cols);
	    double rectangle_height = (bounding_box_.getHeight()-inter_node_spacing_y*(num_rows+1))/(num_rows);
	    
	    // create small rectangles corr. to data nodes
	    double curr_grid_pos_x = bounding_box_ltx + inter_node_spacing_x;
	    double curr_grid_pos_y = bounding_box_lty + inter_node_spacing_y;
	    
	    int curr_row_num = 0; 
	    int curr_col_num = 0;
	    int curr_node_num = _start_data_node_num;
	    
	    boolean possibly_marked0 = false;
	    boolean possibly_marked1 = false;
	    boolean possibly_marked2 = false;
	    if(click_active_) {
		DataNode clicked_data_node = datagraph_.getDataNodeNoChildren(clicked_offset_);
		int clicked_node_path = clicked_data_node.getPath();
		DataguideNode clicked_dg_node = dataguide_.getDataguideNode(clicked_node_path);
		String clicked_dg_path = clicked_dg_node.getPath();
		
		String curr_dg_path = curr_dg_node_.getPath();		    
		
		if(clicked_dg_path.equals(curr_dg_path)) {
		    possibly_marked0 = true;
		}
		else if(clicked_dg_path.startsWith(curr_dg_path)) {
		    if(MyFrame._debug_mode > 5) {
			System.out.println("possible  anc: path="+curr_dg_path);
		    }
		    possibly_marked1 = true;
		}
		else if(curr_dg_path.startsWith(clicked_dg_path)) {
		    if(MyFrame._debug_mode > 5) {
			System.out.println("possible desc: path="+curr_dg_path);
		    }
		    possibly_marked2 = true;
		}
	    }
	    
	    while((curr_node_num - _start_data_node_num) < _num_data_nodes_in_group) {
		if(MyFrame._debug_mode > 7) {
		    System.out.println("Drawing rectangle in grid pos [" + curr_row_num + "," + curr_col_num +"]"); // debug
		}
		if(MyFrame._debug_mode > 8) {
		    System.out.println("Top left corner = (" + curr_grid_pos_x + "," + curr_grid_pos_y +")"); // debug
		}
		
		// bounding box for the data node
		ZRectangle data_bbox = new ZRectangle(curr_grid_pos_x, curr_grid_pos_y, rectangle_width, rectangle_height);
		data_bbox.setFillColor(Color.white);
		
		//ZPolyline data_bbox = new ZPolyline();
		//data_bbox.add(curr_grid_pos_x, curr_grid_pos_y);
		//data_bbox.add(curr_grid_pos_x+rectangle_width, curr_grid_pos_y);
		//data_bbox.add(curr_grid_pos_x+rectangle_width, curr_grid_pos_y+rectangle_height);
		//data_bbox.add(curr_grid_pos_x, curr_grid_pos_y+rectangle_height);
		//data_bbox.add(curr_grid_pos_x, curr_grid_pos_y);
		data_bbox.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
		
		node_.addVisualComponent(data_bbox);
		
		// find current data node
		Long curr_offset = null;
		if(partial_) 
		    curr_offset = (Long)(offsets_vector_.elementAtRandom(curr_node_num));
		else 
		    curr_offset = (Long)(offsets_vector_.elementAt(curr_node_num));
		
		boolean is_marked = false;
		if(click_active_) {
		    if(possibly_marked0) {
			if(clicked_offset_ == curr_offset.longValue()) {
			    is_marked = true;
			    possibly_marked0 = false;
			}
		    }
		    else if(possibly_marked1)
			is_marked = datagraph_.is_ancestor(curr_offset.longValue(), clicked_offset_);
		    else if(possibly_marked2)
			is_marked = datagraph_.is_ancestor(clicked_offset_, curr_offset.longValue());		
		}
		
		
		if(is_marked) { 
		    Color c = new Color(255, 255, 102); // dirty yellow
		    data_bbox.setFillColor(c);
		}
		else {
		    Color c = new Color(51, 255, 255); // fluorescent blue
		    data_bbox.setFillColor(c);
		    //data_bbox.setPenColor(Color.cyan);
		}
		node_.addDataBox(data_bbox, curr_offset);
		
		// create small rectangle for current data node in current grid position 
		
		Rectangle2D bounds = new Rectangle2D.Double(curr_grid_pos_x, curr_grid_pos_y,
							    rectangle_width, rectangle_height); 	    
		MyDataNode curr_visual_data_box = null;
		curr_visual_data_box = new MyDataNode(my_canvas_, node_, sticky_node_, 
						      datagraph_, curr_offset, bounds, 
						      font_size_, camera_mag_);
		
		// find next grid position
		curr_col_num++;
		curr_grid_pos_x += (rectangle_width + inter_node_spacing_x);
		if(curr_col_num == num_cols) {
		    curr_col_num = 0; 
		    curr_grid_pos_x = bounding_box_ltx + inter_node_spacing_x;
		    curr_row_num++; 
		    curr_grid_pos_y += (rectangle_height + inter_node_spacing_y);
		}
		curr_node_num++;
	    }

	    // create prev rect and label
	    Font prev_node_label_font = new Font(null, Font.PLAIN, font_size_); 
	    ZLabel prev_node_label = new ZLabel(new String("<"), prev_node_label_font);
	    ZBounds prev_node_label_bounds = prev_node_label.getBounds();
	    double prev_node_label_width = prev_node_label_bounds.getWidth();
	    double prev_node_label_height = prev_node_label_bounds.getHeight();
	    
	    double ptr_box_top_left_x = bounding_box_ltx;
	    double ptr_box_top_left_y = bounding_box_lty+bounding_box_.getHeight();
	    double ptr_box_width = prev_node_label_width*2; //NODE_WIDTH/PTR_BOX_NODE_WIDTH_RATIO/camera_mag_;
	    double ptr_box_height = prev_node_label_height; //NODE_HEIGHT/PTR_BOX_NODE_HEIGHT_RATIO/camera_mag_;
	    ZRectangle prev_ptr_box = new ZRectangle(ptr_box_top_left_x, ptr_box_top_left_y,
						     ptr_box_width, ptr_box_height);
	    prev_ptr_box.setPenColor(Color.black);
	    prev_ptr_box.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    if(_identifier == 0) {
		prev_ptr_box.setFillColor(Color.white);
		node_.addPrevBox(new ZRectangle(), _dg_node_id);
	    }
	    else {
		prev_ptr_box.setFillColor(Color.lightGray);
		node_.addPrevBox(prev_ptr_box, _dg_node_id);
	    }
	    node_.addVisualComponent(prev_ptr_box);
	    
	    Point2D location = new Point2D.Double(ptr_box_top_left_x+ptr_box_width/2.0f-prev_node_label_width/2.0f,
						  ptr_box_top_left_y+ptr_box_height/2.0f-prev_node_label_height/2.0f);
	    prev_node_label.setPenColor(Color.black);
	    prev_node_label.setTranslation(location);
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Location of prev node label in normal node coords=("+location.getX()+","+location.getY()+")");
	    }
	    //my_canvas_.getCamera().localToCamera(location, node_);
	    //if(MyFrame._debug_mode > 8) {
	    //System.out.println("Location of prev node label in camera coords=("+location.getX()+","+location.getY()+")");
	    //}
	    //my_canvas_.getCamera().cameraToLocal(location, sticky_node_);
	    //if(MyFrame._debug_mode > 8) {
	    //System.out.println("Location of prev node label in sticky node coords=("+location.getX()+","+location.getY()+")");
	    //}
	    //prev_node_label.setTranslation(location);
	    //sticky_node_.addVisualComponent(prev_node_label);
	    node_.addVisualComponent(prev_node_label);

	    // create next rect and label
	    Font next_node_label_font = new Font(null, Font.PLAIN, font_size_); 
	    ZLabel next_node_label = new ZLabel(new String(">"), next_node_label_font);
	    ZBounds next_node_label_bounds = next_node_label.getBounds();
	    double next_node_label_width = next_node_label_bounds.getWidth();
	    double next_node_label_height = next_node_label_bounds.getHeight();
	    ptr_box_width = next_node_label_width*2;
	    ptr_box_height = next_node_label_height;
	    
	    ptr_box_top_left_x = bounding_box_ltx + bounding_box_.getWidth()-ptr_box_width;
	    ptr_box_top_left_y = bounding_box_lty+bounding_box_.getHeight();
	    ZRectangle next_ptr_box = new ZRectangle(ptr_box_top_left_x, ptr_box_top_left_y,
						     ptr_box_width, ptr_box_height);
	    next_ptr_box.setPenColor(Color.black);
	    next_ptr_box.setPenWidth(ABS_PEN_WIDTH/camera_mag_);
	    if(num_rem_data_nodes == 0) {
		next_ptr_box.setFillColor(Color.white);
		node_.addNextBox(new ZRectangle(), _dg_node_id);
	    }
	    else {
		next_ptr_box.setFillColor(Color.lightGray);
		node_.addNextBox(next_ptr_box, _dg_node_id);
	    }
	    node_.addVisualComponent(next_ptr_box);
	
	    location = new Point2D.Double(ptr_box_top_left_x+ptr_box_width/2.0f-next_node_label_width/2.0f,
					  ptr_box_top_left_y+ptr_box_height/2.0f-next_node_label_height/2.0f);
	    next_node_label.setPenColor(Color.black);
	    next_node_label.setTranslation(location);
	    if(MyFrame._debug_mode > 8) {
		System.out.println("Location of next node label in normal node coords=("+location.getX()+","+location.getY()+")");
	    }
	    //my_canvas_.getCamera().localToCamera(location, node_);
	    //if(MyFrame._debug_mode > 8) {
	    //System.out.println("Location of next node label in camera coords=("+location.getX()+","+location.getY()+")");
	    //}
	    //my_canvas_.getCamera().cameraToLocal(location, sticky_node_);
	    //if(MyFrame._debug_mode > 8) {
	    //System.out.println("Location of next node label in sticky node coords=("+location.getX()+","+location.getY()+")");
	    //}
	    //next_node_label.setTranslation(location);
	    //sticky_node_.addVisualComponent(next_node_label);
	    node_.addVisualComponent(next_node_label);	    
	}

	// create label for this group node
	Font node_label_font = new Font(null, Font.PLAIN, font_size_); 
	ZLabel node_label = new ZLabel(node_label_string_, node_label_font);
	ZBounds node_label_bounds = node_label.getBounds();
	double node_label_width = node_label_bounds.getWidth();///camera_mag_;
	double node_label_height = node_label_bounds.getHeight();///camera_mag_;
	Point2D location = new Point2D.Double(bounding_box_ltx+bounding_box_.getWidth()/2.0f-node_label_width/2.0f,
					      bounding_box_lty-node_label_height);
	node_label.setPenColor(Color.black);
	node_label.setTranslation(location);
	if(MyFrame._debug_mode > 8) {
	    System.out.println("Location of node label in normal node coords=("+location.getX()+","+location.getY()+")");
	}
	//my_canvas_.getCamera().localToCamera(location, node_);
	//if(MyFrame._debug_mode > 8) {
	//System.out.println("Location of node label in camera coords=("+location.getX()+","+location.getY()+")");
	//}
	//my_canvas_.getCamera().cameraToLocal(location, sticky_node_);
	//if(MyFrame._debug_mode > 8) {
	//System.out.println("Location of node label in sticky node coords=("+location.getX()+","+location.getY()+")");
	//}
	//node_label.setTranslation(location);
	//sticky_node_.addVisualComponent(node_label);	
	node_.addVisualComponent(node_label);	
    }
}





